@@ -1,15 +1,10 @@
 use std::io::{stderr, stdout, Write};
 
 use anyhow::{anyhow, Result};
-use reqwest::Method;
+use reqwest::{tls, Method};
 
-use crate::{
-    cli::{Cli, Verify},
-    request_items::{
-        Body, RequestItem, RequestItems, FORM_CONTENT_TYPE, JSON_ACCEPT, JSON_CONTENT_TYPE,
-    },
-    url::construct_url,
-};
+use crate::cli::{AuthType, Cli, HttpVersion, Verify};
+use crate::request_items::{Body, RequestItem, FORM_CONTENT_TYPE, JSON_ACCEPT, JSON_CONTENT_TYPE};
 
 pub fn print_curl_translation(args: Cli) -> Result<()> {
     let cmd = translate(args)?;
@@ -58,12 +53,12 @@ impl Command {
         self.push(format!("{}: {}", name, value));
     }
 
-    fn env(&mut self, var: &'static str, value: String) {
-        self.env.push((var, value));
+    fn env(&mut self, var: &'static str, value: impl Into<String>) {
+        self.env.push((var, value.into()));
     }
 
-    fn warn(&mut self, message: String) {
-        self.warnings.push(message);
+    fn warn(&mut self, message: impl Into<String>) {
+        self.warnings.push(message.into());
     }
 }
 
@@ -92,10 +87,7 @@ impl std::fmt::Display for Command {
 }
 
 pub fn translate(args: Cli) -> Result<Command> {
-    let request_items = RequestItems::new(args.request_items);
-    let query = request_items.query();
-    let (headers, headers_to_unset) = request_items.headers()?;
-    let url = construct_url(&args.url, args.default_scheme.as_deref(), query)?;
+    let (headers, headers_to_unset) = args.request_items.headers()?;
 
     let mut cmd = Command::new(args.curl_long);
 
@@ -129,7 +121,9 @@ pub fn translate(args: Cli) -> Result<Command> {
         // showing up right away
         cmd.flag("-N", "--no-buffer");
     }
-    if args.check_status {
+    // Since --fail is bit disruptive than HTTPie's --check-status flag, we will not enable
+    // it unless the user explicitely sets the latter flag
+    if args.check_status.unwrap_or(false) {
         // Suppresses output on failure, unlike us
         cmd.flag("-f", "--fail");
     }
@@ -174,21 +168,53 @@ pub fn translate(args: Cli) -> Result<Command> {
         cmd.push("--key");
         cmd.push(keyfile.to_string_lossy());
     }
+    if let Some(Some(tls_version)) = args.ssl {
+        match tls_version {
+            tls::Version::TLS_1_0 => {
+                cmd.push("--tlsv1.0");
+                cmd.push("--tls-max");
+                cmd.push("1.0");
+            }
+            tls::Version::TLS_1_1 => {
+                cmd.push("--tlsv1.1");
+                cmd.push("--tls-max");
+                cmd.push("1.1");
+            }
+            tls::Version::TLS_1_2 => {
+                cmd.push("--tlsv1.2");
+                cmd.push("--tls-max");
+                cmd.push("1.2");
+            }
+            tls::Version::TLS_1_3 => {
+                cmd.push("--tlsv1.3");
+                cmd.push("--tls-max");
+                cmd.push("1.3");
+            }
+            _ => unreachable!(),
+        }
+    }
     for proxy in args.proxy {
         match proxy {
             crate::cli::Proxy::All(proxy) => {
                 cmd.flag("-x", "--proxy");
-                cmd.push(proxy.into_string());
+                cmd.push(proxy);
             }
             crate::cli::Proxy::Http(proxy) => {
                 // These don't seem to have corresponding flags
-                cmd.env("http_proxy", proxy.into_string());
+                cmd.env("http_proxy", proxy);
             }
             crate::cli::Proxy::Https(proxy) => {
-                cmd.env("https_proxy", proxy.into_string());
+                cmd.env("https_proxy", proxy);
             }
         }
     }
+    if let Some(http_version) = args.http_version {
+        match http_version {
+            HttpVersion::Http10 => cmd.push("--http1.0"),
+            HttpVersion::Http11 => cmd.push("--http1.1"),
+            HttpVersion::Http2 => cmd.push("--http2"),
+        }
+    }
 
     if args.method == Some(Method::HEAD) {
         cmd.flag("-I", "--head");
@@ -203,10 +229,11 @@ pub fn translate(args: Cli) -> Result<Command> {
         // But this is a hack that actually fails if data is sent.
         // See discussion on https://lornajane.net/posts/2014/view-only-headers-with-curl
 
-        let request_type = args.request_type; // For the borrow checker
-        let method = args
-            .method
-            .unwrap_or_else(|| request_items.pick_method(request_type));
+        let method = match args.method {
+            Some(method) => method,
+            // unwrap_or_else causes borrowing issues
+            None => args.request_items.pick_method(),
+        };
         cmd.flag("-I", "--head");
         cmd.flag("-X", "--request");
         cmd.push(method.to_string());
@@ -223,7 +250,7 @@ pub fn translate(args: Cli) -> Result<Command> {
     // We assume that curl's automatic detection of when to do a POST matches
     // ours so we can ignore the None case
 
-    cmd.push(url.to_string());
+    cmd.push(args.url.to_string());
 
     // Payload
     for (header, value) in headers.iter() {
@@ -239,19 +266,30 @@ pub fn translate(args: Cli) -> Result<Command> {
         cmd.push(format!("{}:", header));
     }
     if let Some(auth) = args.auth {
-        // curl implements this flag the same way, including password prompt
-        cmd.flag("-u", "--user");
-        cmd.push(auth);
-    }
-    if let Some(token) = args.bearer {
-        cmd.push("--oauth2-bearer");
-        cmd.push(token);
+        match args.auth_type.unwrap_or_default() {
+            AuthType::basic => {
+                cmd.push("--basic");
+                // curl implements this flag the same way, including password prompt
+                cmd.flag("-u", "--user");
+                cmd.push(auth);
+            }
+            AuthType::digest => {
+                cmd.push("--digest");
+                // curl implements this flag the same way, including password prompt
+                cmd.flag("-u", "--user");
+                cmd.push(auth);
+            }
+            AuthType::bearer => {
+                cmd.push("--oauth2-bearer");
+                cmd.push(auth);
+            }
+        }
     }
 
-    if request_items.is_multipart(args.request_type) {
+    if args.request_items.is_multipart() {
         // We can't use .body() here because we can't look inside the multipart
         // form after construction and we don't want to actually read the files
-        for item in request_items.0 {
+        for item in args.request_items.items {
             match item {
                 RequestItem::JsonField(..) | RequestItem::JsonFieldFromFile(..) => {
                     return Err(anyhow!("JSON values are not supported in multipart fields"));
@@ -268,13 +306,19 @@ pub fn translate(args: Cli) -> Result<Command> {
                     key,
                     file_name,
                     file_type,
+                    file_name_header,
                 } => {
                     cmd.flag("-F", "--form");
+                    let mut val = format!("{}=@{}", key, file_name);
                     if let Some(file_type) = file_type {
-                        cmd.push(format!("{}=@{};type={}", key, file_name, file_type));
-                    } else {
-                        cmd.push(format!("{}=@{}", key, file_name));
+                        val.push_str(";type=");
+                        val.push_str(&file_type);
                     }
+                    if let Some(file_name_header) = file_name_header {
+                        val.push_str(";filename=");
+                        val.push_str(&file_name_header);
+                    }
+                    cmd.push(val);
                 }
                 RequestItem::HttpHeader(..) => {}
                 RequestItem::HttpHeaderToUnset(..) => {}
@@ -282,7 +326,7 @@ pub fn translate(args: Cli) -> Result<Command> {
             }
         }
     } else {
-        match request_items.body(args.request_type)? {
+        match args.request_items.body()? {
             Body::Form(items) => {
                 if items.is_empty() {
                     // Force the header
@@ -318,6 +362,7 @@ pub fn translate(args: Cli) -> Result<Command> {
             Body::File {
                 file_name,
                 file_type,
+                file_name_header: _,
             } => {
                 if let Some(file_type) = file_type {
                     cmd.header("content-type", file_type.to_str()?);
@@ -418,18 +463,18 @@ mod tests {
             ),
             (
                 "xh -d httpbin.org/get",
-                "curl -L -O 'http://httpbin.org/get'",
-                "curl -L -O http://httpbin.org/get",
+                "curl -f -L -O 'http://httpbin.org/get'",
+                "curl -f -L -O http://httpbin.org/get",
             ),
             (
                 "xh -d -o foobar --continue httpbin.org/get",
-                "curl -L -o foobar -C - 'http://httpbin.org/get'",
-                "curl -L -o foobar -C - http://httpbin.org/get",
+                "curl -f -L -o foobar -C - 'http://httpbin.org/get'",
+                "curl -f -L -o foobar -C - http://httpbin.org/get",
             ),
             (
                 "xh --curl-long -d -o foobar --continue httpbin.org/get",
-                "curl --location --output foobar --continue-at - 'http://httpbin.org/get'",
-                "curl --location --output foobar --continue-at - http://httpbin.org/get",
+                "curl --fail --location --output foobar --continue-at - 'http://httpbin.org/get'",
+                "curl --fail --location --output foobar --continue-at - http://httpbin.org/get",
             ),
             (
                 "xh httpbin.org/post @foo.txt",
